#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015-2017, Tomas Tinoco De Rubira.    #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

import os
import csv
from os import listdir
from os.path import join

DIR_CASES = join('.', 'tests', 'resources', 'cases')
DIR_PFSOL = join('.', 'tests', 'resources', 'pf_solutions')

test_cases = [join(DIR_CASES, f) for f in listdir(DIR_CASES)]
test_cases.sort()

def get_pf_solution_file(case_file, dir, sol):
    """
    Gets power flow solution file path.

    Parameters
    ----------
    case_file : path to case file (string)
    dir : path to solution file directory (string)
    sol : solution code extension (string)

    Returns
    -------
    sol_file : path to solution file (string)
    """

    return join(dir, case_file.split(os.sep)[-1]+'.'+sol)

def read_pf_solution_file(sol_file):
    """
    Reads contents of power flow solution file.
    
    Parameters
    ----------
    sol_file : path to solution file.

    Returns
    -------
    sol_data : solution data (dictionary)
    """

    try:

        bus_data = {}
        sol_data = {'v_mag_tol': 0.,
                    'v_ang_tol': 0.,
                    'bus_data': bus_data}

        f = open(sol_file)
        reader = csv.reader(f,delimiter=',')
        
        v_mag_tol,v_ang_tol = list(map(float,next(reader)))

        sol_data['v_mag_tol'] = v_mag_tol
        sol_data['v_ang_tol'] = v_ang_tol

        next(reader) # header

        for row in reader:
            bus_number,code,v_mag,v_ang = int(row[0]),int(row[1]),float(row[2]),float(row[3])
            bus_data[bus_number] = {'v_mag': v_mag, # p.u.
                                    'v_ang': v_ang} # degrees
           
        return sol_data

    except IOError:
        return None



#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

import sys
sys.path.append('.')

import pfnet
import gridopt

net = pfnet.Parser(sys.argv[1]).parse(sys.argv[1])

method = gridopt.power_flow.new_method('ACPF')

method.set_parameters({'solver': 'nr', 'quiet': True, 'feastol': 1e-4})

method.solve(net)

results = method.get_results()

print((results['solver status']))

print((results['solver iterations']))

print((results['network snapshot'].bus_v_max))

method.update_network(net)

print(('%.2e %.2e' %(net.bus_P_mis,net.bus_Q_mis)))



#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

class PFmethodError(Exception):
    pass

class PFmethodError_NoProblem(PFmethodError):    
    def __init__(self):
        PFmethodError.__init__(self, 'no problem solved')

class PFmethodError_BadProblem(PFmethodError):    
    def __init__(self):
        PFmethodError.__init__(self, 'error while creating PF problem')

class PFmethodError_BadFlowLimits(PFmethodError):
    def __init__(self):
        PFmethodError.__init__(self, 'invalid flow limits')

class PFmethodError_BadVarLimits(PFmethodError):
    def __init__(self):
        PFmethodError.__init__(self, 'invalid variable limits')

class PFmethodError_BadParams(PFmethodError):
    def __init__(self, keys):
        PFmethodError.__init__(self, 'invalid method parameters: {}'.format(keys))

class PFmethodError_BadParamValue(PFmethodError):
    def __init__(self, key):
        PFmethodError.__init__(self, 'invalid value for parameter {}'.format(key))

class PFmethodError_BadOptSolver(PFmethodError):
    def __init__(self, param=''):
        PFmethodError.__init__(self, 'invalid optimization solver %s' %param)

class PFmethodError_ParamNotBool(PFmethodError):
    def __init__(self):
        PFmethodError.__init__(self, 'parameter value must be True or False')

class PFmethodError_TranVReg(PFmethodError):
        def __init__(self, msg):
            PFmethodError.__init__(self, 'error in transformer voltage regulation :%s' %msg)

class PFmethodError_ShuntVReg(PFmethodError):
        def __init__(self, msg):
            PFmethodError.__init__(self, 'error in shunt voltage regulation: %s' %msg)

class PFmethodError_SolverError(PFmethodError):
    def __init__(self, msg):
        PFmethodError.__init__(self, msg)





#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

from __future__ import print_function
import time
import numpy as np
from .method_error import *
from .method import PFmethod

class DCPF(PFmethod):
    """
    DC power flow method.
    """

    name = 'DCPF'
    
    _parameters = {'quiet': False,
                   'solver' : 'superlu'}
    
    def __init__(self):

        PFmethod.__init__(self)

        self._parameters = DCPF._parameters.copy()
        self._parameters['solver_parameters'] = {'superlu': {},
                                                 'mumps': {}}

    def create_problem(self,net):

        import pfnet

        # Parameters
        params = self._parameters
        
        # Clear flags
        net.clear_flags()

        # Voltages angles (not slack)
        net.set_flags('bus',
                      'variable',
                      'not slack',
                      'voltage angle')

        # Gen active powers (slack)
        net.set_flags('generator',
                      'variable',
                      'slack',
                      'active power')

        # Check
        try:
            assert(net.num_vars == net.num_buses-1+net.get_num_slack_gens())
        except AssertionError:
            raise PFmethodError_BadProblem()

        # Set up problem
        problem = pfnet.Problem(net)
        problem.add_constraint(pfnet.Constraint('DC power balance',net))
        problem.add_constraint(pfnet.Constraint('generator active power participation',net))
        problem.analyze()

        # Return
        return problem
                    
    def solve(self,net):

        from optalg.lin_solver import new_linsolver
        
        # Parameters
        params = self._parameters
        solver_name = params['solver']

        # Copy network
        net = net.get_copy(merge_buses=True)
        
        # Problem
        t0 = time.time()
        problem = self.create_problem(net)
        problem_time = time.time()-t0
        
        A = problem.A
        b = problem.b
        x = problem.x

        # Solve
        update = True
        t0 = time.time()
        try:
            assert(A.shape[0] == A.shape[1])
            linsolver = new_linsolver(solver_name,'unsymmetric')
            linsolver.analyze(A)
            x = linsolver.factorize_and_solve(A,b)
        except Exception as e:
            update = False
            raise PFmethodError_SolverError(e)
        finally:
            
            # Update network
            if update:
                net.set_var_values(x)
                net.update_properties()
                net.clear_sensitivities()

            # Save results
            self.set_solver_name(solver_name)
            self.set_solver_status('solved' if update else 'error')
            self.set_solver_message('')
            self.set_solver_iterations(1)
            self.set_solver_time(time.time()-t0)
            self.set_solver_primal_variables(x)
            self.set_solver_dual_variables(4*[None])
            self.set_problem(None) # skip for now
            self.set_problem_time(problem_time)
            self.set_network_snapshot(net)



#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

from . import power_flow
from . import scripts

from .version import __version__



#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

from __future__ import print_function
import time
import numpy as np
from .method_error import *
from .method import PFmethod
from numpy.linalg import norm

class ACPF(PFmethod):
    """
    AC power flow method.
    """
    
    CONTROL_MODE_LOCKED = 'locked'
    CONTROL_MODE_FREE = 'free'
    CONTROL_MODE_REG = 'regulating'

    name = 'ACPF'
    
    _parameters = {'weight_vmag': 1e0,       # weight for voltage magnitude regularization
                   'weight_vang': 1e-3,      # weight for angle difference regularization
                   'weight_powers': 1e-3,    # weight for gen powers regularization
                   'weight_controls': 1e0,   # weight for control deviation penalty
                   'weight_var': 1e-5,       # weight for general variable regularization
                   'v_min_clip': 0.5,        # lower v threshold for clipping
                   'v_max_clip': 1.5,        # upper v threshold for clipping
                   'v_limits': False,        # voltage magnitude limits
                   'Q_limits': True,         # flag for enforcing generator, VSC and FACTS reactive power limits
                   'Q_mode': 'regulating',   # reactive power mode: free, regulating
                   'shunt_limits': True,     # flag for enforcing switched shunt susceptance limits
                   'shunt_mode': 'locked',   # switched shunts mode: locked, free, regulating
                   'tap_limits': True,       # flag for enforcing transformer tap ratio limits
                   'tap_mode': 'locked',     # transformer tap ratio mode: locked, free, regulating
                   'lock_vsc_P_dc': True,    # flag for locking vsc P dc
                   'lock_csc_P_dc': True,    # flag for locking csc P dc
                   'lock_csc_i_dc': True,    # flag for locking csc i dc
                   'vdep_loads': False,      # flag for modeling voltage dependent loads
                   'pvpq_start_k': 0,        # start iteration number for PVPQ switching heuristics
                   'vmin_thresh': 0.1,       # minimum voltage magnitude threshold
                   'gens_redispatch': False, # flag for allowing active power redispatch
                   'shunts_round': True,     # flag for rounding discrete switched shunt susceptances (not supported yet)
                   'taps_round': True,       # flag for rounding discrete transformer tap ratios (not supported yet)
                   'v_mag_warm_ref': False,  # flag for using current v_mag as reference in v_mag regularization
                   'solver': 'nr',           # OPTALG optimization solver: augl, ipopt, nr, inlp
                   'tap_step': 0.5,          # tap ratio acceleration factor (NR only)
                   'shunt_step': 0.5,        # susceptance acceleration factor (NR only)
                   'dtap': 1e-4,             # tap ratio perturbation (NR only)
                   'dsus': 1e-4}             # susceptance perturbation (NR only)

    _parameters_augl = {'feastol' : 1e-4,
                        'optol' : 1e0,
                        'kappa' : 1e-5,
                        'theta_max': 1e-6,
                        'sigma_init_max': 1e9}

    _parameters_ipopt = {}

    _parameters_inlp = {'feastol' : 1e-4,
                        'optol' : 1e0}
    
    _parameters_nr = {}
                  
    def __init__(self):

        from optalg.opt_solver import OptSolverAugL, OptSolverIpopt, OptSolverNR, OptSolverINLP

        # Parent init
        PFmethod.__init__(self)

        # Solver params
        augl_params = OptSolverAugL.parameters.copy()
        augl_params.update(self._parameters_augl)   # overwrite defaults

        ipopt_params = OptSolverIpopt.parameters.copy()
        ipopt_params.update(self._parameters_ipopt) # overwrite defaults

        inlp_params = OptSolverINLP.parameters.copy()
        inlp_params.update(self._parameters_inlp)   # overwrite defaults

        nr_params = OptSolverNR.parameters.copy()
        nr_params.update(self._parameters_nr)       # overwrite defaults

        self._parameters = ACPF._parameters.copy()
        self._parameters['solver_parameters'] = {'augl': augl_params,
                                                 'ipopt': ipopt_params,
                                                 'nr': nr_params,
                                                 'inlp': inlp_params}

    def create_problem(self, net):

        solver_name = self._parameters['solver']

        if solver_name == 'nr':
            return self.create_problem_nr(net)
        else:
            return self.create_problem_opt(net)

    def create_problem_nr(self, net):

        import pfnet

        # Parameters
        params = self._parameters
        Q_mode = params['Q_mode']
        Q_limits = params['Q_limits']
        shunt_mode = params['shunt_mode']
        shunt_limits = params['shunt_limits']
        tap_mode = params['tap_mode']
        tap_limits = params['tap_limits']
        lock_vsc_P_dc = params['lock_vsc_P_dc']
        lock_csc_P_dc = params['lock_csc_P_dc']
        lock_csc_i_dc = params['lock_csc_i_dc']
        vdep_loads = params['vdep_loads']
        gens_redispatch = params['gens_redispatch']

        # Check shunt options 
        if shunt_mode not in [self.CONTROL_MODE_LOCKED,
                               self.CONTROL_MODE_REG]:
            raise ValueError('invalid shunts mode')
        if shunt_mode == self.CONTROL_MODE_REG and not shunt_limits:
            raise ValueError('unsupported shunts configuration')

        # Check tap options 
        if tap_mode not in [self.CONTROL_MODE_LOCKED,
                               self.CONTROL_MODE_REG]:
            raise ValueError('invalid taps mode')
        if tap_mode == self.CONTROL_MODE_REG and not tap_limits:
            raise ValueError('unsupported taps configuration')
        
        # Check Q options
        if Q_mode != self.CONTROL_MODE_REG:
            raise ValueError('invalid reactive power mode')
        
        # Check other options
        if gens_redispatch:
            raise ValueError('generation redispatch not supported')
        if not lock_vsc_P_dc:
            raise ValueError('VSC P DC must be locked')
        if not lock_csc_P_dc:
            raise ValueError('CSC P DC must be locked')
        if not lock_csc_i_dc:
            raise ValueError('CSC i DC must be locked')
        
        # Clear flags
        net.clear_flags()

        # Buses
        net.set_flags('bus',
                      'variable',
                      'not slack',
                      'voltage angle')
        net.set_flags('bus',
                      'variable',
                      'any',
                      'voltage magnitude')

        # Generators
        net.set_flags('generator',
                      'variable',
                      'slack',
                      'active power')
        
        net.set_flags('generator',
                      'variable',
                      'regulator',
                      'reactive power')
        
        # VSC HVDC
        net.set_flags('vsc converter',
                      'variable',
                      'any',
                      ['dc power', 'active power', 'reactive power'])
        
        # CSC HVDC
        net.set_flags('csc converter',
                      'variable',
                      'any',
                      ['dc power', 'active power', 'reactive power'])
        
        # DC buses
        net.set_flags('dc bus',
                      'variable',
                      'any',
                      'voltage')
        
        # FACTS
        net.set_flags('facts',
                      'variable',
                      'any',
                      'all')
        
        # Loads
        if vdep_loads:
            for load in net.loads:
                if load.is_voltage_dependent() and load.is_in_service():
                    net.set_flags_of_component(load,
                                               'variable',
                                               ['active power', 'reactive power'])
                    
        # Tap changers 
        if tap_mode != self.CONTROL_MODE_LOCKED:
            net.set_flags('branch',
                          ['variable', 'fixed'],
                          'tap changer - v',
                          'tap ratio')
        
        # Switched shunts
        if shunt_mode != self.CONTROL_MODE_LOCKED:
            net.set_flags('shunt',
                          ['variable', 'fixed'],
                          'switching - v',
                          'susceptance')
        
        # Set up problem
        problem = pfnet.Problem(net)
        
        problem.add_constraint(pfnet.Constraint('AC power balance', net))
        problem.add_constraint(pfnet.Constraint('HVDC power balance', net))
        problem.add_constraint(pfnet.Constraint('generator active power participation', net))
        problem.add_constraint(pfnet.Constraint('variable fixing', net))
        problem.add_constraint(pfnet.Constraint('VSC converter equations', net))
        problem.add_constraint(pfnet.Constraint('CSC converter equations', net))
        problem.add_constraint(pfnet.Constraint('FACTS equations', net))
        problem.add_constraint(pfnet.Constraint('VSC DC voltage control', net))
        problem.add_constraint(pfnet.Constraint('CSC DC voltage control', net))
        problem.add_constraint(pfnet.Constraint('VSC DC power control', net))
        problem.add_constraint(pfnet.Constraint('CSC DC power control', net))
        problem.add_constraint(pfnet.Constraint('CSC DC current control', net))

        problem.add_constraint(pfnet.Constraint('PVPQ switching', net))
        problem.add_constraint(pfnet.Constraint('switching power factor regulation', net))
        problem.add_constraint(pfnet.Constraint('switching FACTS active power control', net))
        problem.add_constraint(pfnet.Constraint('switching FACTS reactive power control', net))
        
        if vdep_loads:
            problem.add_constraint(pfnet.Constraint('load voltage dependence', net))

        if Q_limits:            
            problem.add_heuristic(pfnet.Heuristic('PVPQ switching', net))
            problem.add_heuristic(pfnet.Heuristic('switching power factor regulation', net))

        problem.analyze()
        
        # Check
        if (problem.J.shape[0] + problem.A.shape[0] != problem.get_num_primal_variables()):
            raise PFmethodError_BadProblem()
        
        # Return
        return problem

    def create_problem_opt(self, net):

        import pfnet

        # Parameters
        params = self._parameters
        wm = params['weight_vmag']
        wa = params['weight_vang']
        wp = params['weight_powers']
        wc = params['weight_controls']
        wv = params['weight_var']
        v_limits = params['v_limits']
        Q_mode = params['Q_mode']
        Q_limits = params['Q_limits']
        shunt_mode = params['shunt_mode']
        shunt_limits = params['shunt_limits']
        tap_mode = params['tap_mode']
        tap_limits = params['tap_limits']
        lock_vsc_P_dc = params['lock_vsc_P_dc']
        lock_csc_P_dc = params['lock_csc_P_dc']
        lock_csc_i_dc = params['lock_csc_i_dc']
        vdep_loads = params['vdep_loads']
        v_mag_warm_ref = params['v_mag_warm_ref']
        gens_redispatch = params['gens_redispatch']
        
        # Check shunt options
        if shunt_mode not in [self.CONTROL_MODE_LOCKED,
                               self.CONTROL_MODE_FREE,
                               self.CONTROL_MODE_REG]:
            raise ValueError('invalid shunts mode')
        if shunt_mode == self.CONTROL_MODE_REG and not shunt_limits:
            raise ValueError('unsupported shunts configuration')

        # Check tap options
        if tap_mode not in [self.CONTROL_MODE_LOCKED,
                             self.CONTROL_MODE_FREE,
                             self.CONTROL_MODE_REG]:
            raise ValueError('invalid taps mode')
        if tap_mode == self.CONTROL_MODE_REG and not tap_limits:
            raise ValueError('unsupported taps configuration')

        # Check Q options
        if Q_mode not in [self.CONTROL_MODE_REG,
                          self.CONTROL_MODE_FREE]:
            raise ValueError('invalid reactive power mode')

        # Clear flags
        net.clear_flags()

        # Buses
        net.set_flags('bus',
                      'variable',
                      'not slack',
                      'voltage angle')
        net.set_flags('bus',
                      'variable',
                      'any',
                      'voltage magnitude')
        if Q_mode == self.CONTROL_MODE_REG and not Q_limits: 
            net.set_flags('bus',
                          'fixed',
                          'v set regulated',
                          'voltage magnitude')
        if v_limits:
            net.set_flags('bus',
                          'bounded',
                          'any',
                          'voltage magnitude')
            
        # Genertors 
        if gens_redispatch:    
            net.set_flags('generator',
                          ['variable', 'bounded'],
                          'any',
                          'active power')
        else:  
            net.set_flags('generator',
                          'variable',
                          'slack',
                          'active power')
        net.set_flags('generator',
                      'variable',
                      'regulator',
                      'reactive power')
        if Q_mode == self.CONTROL_MODE_FREE and Q_limits: 
            net.set_flags('generator',
                          'bounded',
                          'regulator',
                          'reactive power')
            
        # Loads
        if vdep_loads:
            for load in net.loads:
                if load.is_voltage_dependent() and load.is_in_service():
                    net.set_flags_of_component(load,
                                               'variable',
                                               ['active power', 'reactive power'])

        # VSC HVDC
        net.set_flags('vsc converter',
                      'variable',
                      'any',
                      ['dc power', 'active power', 'reactive power'])
        if Q_mode == self.CONTROL_MODE_FREE and Q_limits: 
            net.set_flags('vsc converter',
                          'bounded',
                          'any',
                          'reactive power')
            
        # CSC HVDC
        net.set_flags('csc converter',
                      'variable',
                      'any',
                      ['dc power', 'active power', 'reactive power'])
            
        # DC buses
        net.set_flags('dc bus',
                      'variable',
                      'any',
                      'voltage')

        # FACTS
        net.set_flags('facts',
                      'variable',
                      'any',
                      'all')
        if Q_mode == self.CONTROL_MODE_FREE and Q_limits:        
            net.set_flags('facts',
                          'bounded',
                          'any',
                          'reactive power')
            
        # Tap changers 
        if tap_mode != self.CONTROL_MODE_LOCKED:
            net.set_flags('branch', 
                          'variable',
                          'tap changer - v',
                          'tap ratio')
        if tap_mode == self.CONTROL_MODE_FREE and tap_limits:
            net.set_flags('branch',
                          'bounded',
                          'tap changer - v',
                          'tap ratio')
            
        # Swtiched shunts
        if shunt_mode != self.CONTROL_MODE_LOCKED:
            net.set_flags('shunt', 
                          'variable',
                          'switching - v',
                          'susceptance')
        if shunt_mode == self.CONTROL_MODE_FREE and shunt_limits:
            net.set_flags('shunt',
                          'bounded',
                          'switching - v',
                          'susceptance')
            
        # Set up problem
        problem = pfnet.Problem(net)
        
        problem.add_constraint(pfnet.Constraint('AC power balance', net))
        problem.add_constraint(pfnet.Constraint('HVDC power balance', net))
        problem.add_constraint(pfnet.Constraint('generator active power participation', net))
        problem.add_constraint(pfnet.Constraint('VSC converter equations', net))
        problem.add_constraint(pfnet.Constraint('CSC converter equations', net))
        problem.add_constraint(pfnet.Constraint('FACTS equations', net))
        problem.add_constraint(pfnet.Constraint('VSC DC voltage control', net))
        problem.add_constraint(pfnet.Constraint('CSC DC voltage control', net))
        problem.add_constraint(pfnet.Constraint('power factor regulation', net))
        
        if lock_vsc_P_dc:
            problem.add_constraint(pfnet.Constraint('VSC DC power control', net))
        if lock_csc_P_dc:
            problem.add_constraint(pfnet.Constraint('CSC DC power control', net))
        if lock_csc_i_dc:
            problem.add_constraint(pfnet.Constraint('CSC DC current control', net))

        func = pfnet.Function('voltage magnitude regularization', wm/(net.get_num_buses(True)+1.), net)
        func.set_parameter('v_set_reference', not v_mag_warm_ref)
        problem.add_function(func)

        problem.add_function(pfnet.Function('variable regularization', wv/(net.num_vars+1.), net))
        problem.add_function(pfnet.Function('voltage angle regularization', wa/(net.get_num_buses(True)+1.), net))
        problem.add_function(pfnet.Function('generator powers regularization', wp/(net.get_num_generators(True)+1.), net))
        problem.add_function(pfnet.Function('VSC DC power control', wc/(net.get_num_vsc_converters(True)+1.), net))
        problem.add_function(pfnet.Function('CSC DC power control', wc/(net.get_num_csc_converters(True)+1.), net))
        problem.add_function(pfnet.Function('CSC DC current control', wc/(net.get_num_csc_converters(True)+1.), net))
        problem.add_function(pfnet.Function('FACTS active power control', wc/(net.get_num_facts(True)+1.), net))
        problem.add_function(pfnet.Function('FACTS reactive power control', wc/(net.get_num_facts(True)+1.), net))
        
        if gens_redispatch:
            problem.add_function(pfnet.Function('generation redispatch penalty', wc/(net.get_num_generators(True)+1.), net))
            
        if Q_mode == self.CONTROL_MODE_REG and Q_limits: 
            problem.add_constraint(pfnet.Constraint('voltage set point regulation', net))

        if net.num_fixed > 0:
            problem.add_constraint(pfnet.Constraint('variable fixing', net))
            
        if tap_mode != self.CONTROL_MODE_LOCKED:
            problem.add_function(pfnet.Function('tap ratio regularization', wc/(net.get_num_tap_changers_v(True)+1.), net))
            if tap_mode == self.CONTROL_MODE_REG and tap_limits: 
                problem.add_constraint(pfnet.Constraint('voltage regulation by transformers', net))
            
        if shunt_mode != self.CONTROL_MODE_LOCKED:
            problem.add_function(pfnet.Function('susceptance regularization', wc/(net.get_num_switched_v_shunts(True)+1.), net))
            if shunt_mode == self.CONTROL_MODE_REG and shunt_limits:
                problem.add_constraint(pfnet.Constraint('voltage regulation by shunts', net))
            
        if vdep_loads:
            problem.add_constraint(pfnet.Constraint('load voltage dependence', net))
            
        if net.num_bounded > 0: 
            problem.add_constraint(pfnet.Constraint('variable bounds', net))
            
        # Analyze
        problem.analyze()
        
        # Return
        return problem
            
    def solve(self, net, save_problem=False):

        from optalg.opt_solver import OptSolverError, OptTermination, OptCallback
        from optalg.opt_solver import OptSolverAugL, OptSolverIpopt, OptSolverNR, OptSolverINLP
        
        # Parameters
        params = self._parameters
        Q_mode = params['Q_mode']
        shunt_mode = params['shunt_mode']
        shunts_round = params['shunts_round']
        tap_mode = params['tap_mode']
        taps_round = params['taps_round']
        vmin_thresh = params['vmin_thresh']
        solver_name = params['solver']
        solver_params = params['solver_parameters']
        v_min_clip = params['v_min_clip']
        v_max_clip = params['v_max_clip']

        # Opt solver
        if solver_name == 'augl':
            solver = OptSolverAugL()
        elif solver_name == 'ipopt':
            solver = OptSolverIpopt()
        elif solver_name == 'inlp':
            solver = OptSolverINLP()
        elif solver_name == 'nr':
            solver = OptSolverNR()
        else:
            raise PFmethodError_BadOptSolver()
        solver.set_parameters(solver_params[solver_name])

        # Copy network
        net = net.get_copy(merge_buses=True)
        self.set_network_snapshot(net)

        # Clipping
        for bus in net.buses:
            bus.v_mag = np.minimum(np.maximum(bus.v_mag, v_min_clip), v_max_clip)

        # Problem
        t0 = time.time()
        problem = self.create_problem(net)
        problem_time = time.time()-t0

        # Callbacks
        def c1(s):
            if (s.k != 0 and params['tap_limits'] and tap_mode == self.CONTROL_MODE_REG and
                norm(s.problem.f, np.inf) < 100.*solver_params['nr']['feastol']):
                try:
                    self.apply_tran_v_regulation(s)
                except Exception as e:
                    raise PFmethodError_TranVReg(e)
            
        def c2(s):
            if (s.k != 0 and params['shunt_limits'] and shunt_mode == self.CONTROL_MODE_REG and
                norm(s.problem.f, np.inf) < 100.*solver_params['nr']['feastol']):
                try:
                    self.apply_shunt_v_regulation(s)
                except Exception as e:
                    raise PFmethodError_ShuntVReg(e)                

        def c3(s):
            if (s.k >= params['pvpq_start_k'] and params['Q_limits'] and Q_mode == self.CONTROL_MODE_REG):
                prob = s.problem.wrapped_problem
                prob.apply_heuristics(s.x)
                s.problem.A = prob.A
                s.problem.b = prob.b

        if solver_name == 'nr':
            solver.add_callback(OptCallback(c1))
            solver.add_callback(OptCallback(c2))
            solver.add_callback(OptCallback(c3))
                
        # Termination
        def t1(s):
            if np.min(s.problem.wrapped_problem.network.bus_v_min) < vmin_thresh:
                return True
            else:
                return False
        solver.add_termination(OptTermination(t1, 'low voltage'))
            
        # Info printer
        info_printer = self.get_info_printer()
        solver.set_info_printer(info_printer)
        
        # Solve
        update = True
        t0 = time.time()
        try:
            solver.solve(problem)
        except OptSolverError as e:
            raise PFmethodError_SolverError(e)
        except Exception as e:
            update = False
            raise e
        finally:
            
            # Update network
            if update:
                net.set_var_values(solver.get_primal_variables()[:net.num_vars])
                net.update_properties()
                net.clear_sensitivities()
                if solver_name != 'nr':
                    problem.store_sensitivities(*solver.get_dual_variables())

            # Save results
            self.set_solver_name(solver_name)
            self.set_solver_status(solver.get_status())
            self.set_solver_message(solver.get_error_msg())
            self.set_solver_iterations(solver.get_iterations())
            self.set_solver_time(time.time()-t0)
            self.set_solver_primal_variables(solver.get_primal_variables())
            self.set_solver_dual_variables(solver.get_dual_variables())
            self.set_problem(problem if save_problem else None)
            self.set_problem_time(problem_time)
            self.set_network_snapshot(net)
 
    def get_info_printer(self):

        # Parameters
        solver_name = self._parameters['solver']

        # Define
        def info_printer(solver,header):
            net = solver.problem.wrapped_problem.network
            if header:
                print('{0:^5}'.format('vmax'), end=' ')
                print('{0:^5}'.format('vmin'), end=' ')
                print('{0:^8}'.format('gvdev'), end=' ')
                print('{0:^8}'.format('gQvio'))
            else:
                print('{0:^5.2f}'.format(np.average(net.bus_v_max)), end=' ')
                print('{0:^5.2f}'.format(np.average(net.bus_v_min)), end=' ')
                print('{0:^8.1e}'.format(np.average(net.gen_v_dev)), end=' ')
                print('{0:^8.1e}'.format(np.average(net.gen_Q_vio)))

        # Return
        return info_printer

    def apply_shunt_v_regulation(self,solver):

        # Local variables
        dsus = self._parameters['dsus']
        step = self._parameters['shunt_step']
        p = solver.problem.wrapped_problem
        net = p.network
        x = solver.x
        eps = 1e-8

        # Fix constraints
        c = p.find_constraint('variable fixing')
        A = c.A        
        b = c.b
        
        # Rhs
        rhs = np.hstack((np.zeros(p.f.size),np.zeros(p.b.size)))

        # Offset
        offset = 0
        for c in p.constraints:
            if c.name == 'variable fixing':
                break
            else:
                offset += c.A.shape[0]

        # Violation check
        for i in range(net.num_buses):

            bus = net.get_bus(i)

            if bus.is_regulated_by_shunt(True) and not bus.is_slack():
                
                assert(bus.has_flags('variable','voltage magnitude'))

                for t in range(net.num_periods):

                    v = x[bus.index_v_mag[t]]
                    vmax = bus.v_max_reg
                    vmin = bus.v_min_reg

                    assert(len(bus.reg_shunts) > 0)
                    assert(vmax >= vmin)

                    # Violation
                    if v > vmax or v < vmin:

                        for reg_shunt in bus.reg_shunts:

                            if not reg_shunt.is_in_service():
                                continue
                    
                            assert(reg_shunt.has_flags('variable','susceptance'))
                        
                            s = x[reg_shunt.index_b[t]]
                            smax = reg_shunt.b_max
                            smin = reg_shunt.b_min
                            assert(smin <= smax)
                            
                            # Fix constr index
                            k = int(np.where(A.col == reg_shunt.index_b[t])[0])
                            i = A.row[k]
                            
                            assert(np.abs(b[i]-x[reg_shunt.index_b[t]]) < eps)
                            assert(A.data[k] == 1.)
                            
                            # Sensitivity
                            assert(rhs[p.f.size+offset+i] == 0.)
                            rhs[p.f.size+offset+i] = dsus
                            dx = solver.linsolver.solve(rhs)
                            dv = dx[bus.index_v_mag[t]]
                            dvds = dv/dsus
                            rhs[p.f.size+offset+i] = 0.
                            
                            # Adjustment
                            dv = (vmax+vmin)/2.-v
                            ds = step*dv/dvds if dvds != 0. else 0.
                            snew = np.maximum(np.minimum(s+ds,smax),smin)
                            x[reg_shunt.index_b[t]] = snew
                            b[i] = snew
                            if np.abs(snew-s) > eps:
                                break
                            
        # Update
        solver.func(x)
        p.update_lin()
        solver.problem.A = p.A
        solver.problem.b = p.b

    def apply_tran_v_regulation(self,solver):
        
        # Local variables
        dtap = self._parameters['dtap']
        step = self._parameters['tap_step']
        p = solver.problem.wrapped_problem
        net = p.network
        x = solver.x
        eps = 1e-8

        # Fix constraints
        c = p.find_constraint('variable fixing')
        A = c.A        
        b = c.b

        # Rhs
        rhs = np.hstack((np.zeros(p.f.size),np.zeros(p.b.size)))
        
        # Offset
        offset = 0
        for c in p.constraints:
            if c.name == 'variable fixing':
                break
            else:
                offset += c.A.shape[0]

        # Violation check
        for i in range(net.num_buses):

            bus = net.get_bus(i)

            if bus.is_regulated_by_tran(True) and not bus.is_slack():
                
                assert(bus.has_flags('variable','voltage magnitude'))

                for tau in range(net.num_periods):
                    
                    v = x[bus.index_v_mag[tau]]
                    vmax = bus.v_max_reg
                    vmin = bus.v_min_reg
                    
                    assert(len(bus.reg_trans) > 0)
                    assert(vmax > vmin)
                
                    # Violation
                    if v > vmax or v < vmin:
                        
                        for reg_tran in bus.reg_trans:

                            if not reg_tran.is_in_service():
                                continue
                                                        
                            assert(reg_tran.has_flags('variable','tap ratio'))
                            
                            t = x[reg_tran.index_ratio[tau]]
                            tmax = reg_tran.ratio_max
                            tmin = reg_tran.ratio_min
                            assert(tmax >= tmin)
                            
                            # Fix constr index
                            k = int(np.where(A.col == reg_tran.index_ratio[tau])[0])
                            i = A.row[k]
                            assert(np.abs(b[i]-x[reg_tran.index_ratio[tau]]) < eps)
                            assert(A.data[k] == 1.)
                            
                            # Sensitivity
                            assert(rhs[p.f.size+offset+i] == 0.)
                            rhs[p.f.size+offset+i] = dtap
                            dx = solver.linsolver.solve(rhs)
                            dv = dx[bus.index_v_mag[tau]]
                            dvdt = dv/dtap
                            rhs[p.f.size+offset+i] = 0.
                            
                            # Adjustment
                            dv = (vmax+vmin)/2.-v
                            dt = step*dv/dvdt if dvdt != 0. else 0.
                            tnew = np.maximum(np.minimum(t+dt,tmax),tmin)
                            x[reg_tran.index_ratio[tau]] = tnew
                            b[i] = tnew
                            if np.abs(tnew-t) > eps:
                                break

        # Update
        solver.func(x)        
        p.update_lin()
        solver.problem.A = p.A
        solver.problem.b = p.b




#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

from __future__ import print_function
import sys
import pstats
import argparse
import cProfile
from ..power_flow import new_method, PFmethodError

methods = ['ACOPF', 'ACPF', 'DCOPF', 'DCPF']

def create_parser():
    
    # Construct parser
    parser = argparse.ArgumentParser(description='Power grid optimization package.')
    
    # Case
    parser.add_argument('case', type=str,
                        help='filename of power flow case to solve')

    # Method
    parser.add_argument('method', choices=methods,metavar='method',
                        help='PF or OPF method')
    
    # Params
    class ParamAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            for value in values:
                if '=' not in value:
                    raise argparse.ArgumentTypeError("invalid parameter name-value pair")
                n,v = value.split('=')
                params = getattr(namespace, self.dest)
                setattr(namespace, self.dest,params+[(n, v)])
    parser.add_argument('--parameters', action=ParamAction, dest='params', nargs='*', default=[],
                        help='parameter name-value pairs')

    # Profile
    parser.add_argument('--profile', action='store_true', default=False,
                        help='flag for profiling execution')

    # Flat start
    parser.add_argument('--flatstart', action='store_true', default=False,
                        help='flag for flat starting point')

    # Quiet
    parser.add_argument('--quiet', action='store_true', default=False,
                        help='flag for supressing output')
    
    # Write
    parser.add_argument('--write',
                        dest='outfile',
                        type=str,
                        default='',
                        help='name of output file')

    return parser

def main(args=None):

    import pfnet

    parser = create_parser()
    args = parser.parse_args(args=args if args is not None else sys.argv[1:])
    
    try:
        
        # Network
        net = pfnet.Parser(args.case).parse(args.case)
        if not args.quiet:
            net.show_components()
        
        if args.flatstart:

            # Flat start
            for bus in net.buses:
                bus.v_mag = 1
                bus.v_ang = 0
        
        # Method
        method = new_method(args.method)
            
        # Parameters
        method.set_parameters(strparams=dict(args.params))
        if args.quiet:
            method.set_parameters({'quiet': True})
            
        if args.profile:

            # Profile
            cProfile.runctx('method.solve(net)', globals(), locals(), '.prof')
            pstats.Stats('.prof').strip_dirs().sort_stats('cumulative').print_stats(20)
        else:

            # Solve
            try:
                method.solve(net)
                if not args.quiet:
                    print(method.results['solver status'])
            except PFmethodError as e:
                if not args.quiet:
                    print(e)
            
            # Update
            method.update_network(net)
                
            # Show network props
            if not args.quiet:
                net.show_properties()

            # Write
            if args.outfile:
                p = pfnet.Parser(args.outfile)
                p.write(net, args.outfile)

    finally:
        pass
    
# Main function            
if __name__ == "__main__":
    main()



#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

from .dc_pf import DCPF
from .dc_opf import DCOPF
from .ac_pf import ACPF
from .ac_opf import ACOPF
from .method import PFmethod
from .method_error import PFmethodError

methods = [DCPF, DCOPF, ACPF, ACOPF]

def new_method(name):
    """
    Creates a power flow or optimal power flow method.
    
    Parameters
    ----------
    name : {``'DCPF'``, ``'DCOPF'``, ``'ACPF'``, ``'ACOPF'``}
    """
    
    try:
        return methods[list([x.name for x in methods]).index(name)]()
    except ValueError:
        raise ValueError('invalid method name')



#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

import sys
sys.path.append('.')

import pfnet
import gridopt

net = pfnet.Parser(sys.argv[1]).parse(sys.argv[1])

method = gridopt.power_flow.new_method('DCOPF')

method.solve(net)

print((method.results['solver status']))

method.update_network(net)

# generation cost ($/hour)
print((net.gen_P_cost))

# sensitivity to power balance 
bus = net.get_bus(4)
print(("bus %2d %.2e" %(bus.index,bus.sens_P_balance)))

# sensitivity to flow limits 
branch = net.get_branch(6)
print(("branch %2d %.2e %.2e" %(branch.index,
                               branch.sens_P_u_bound,
                               branch.sens_P_l_bound)))

# sensitivity to  gen limits
gen = net.get_generator(2)
print(("gen %2d %.2e %.2e" %(gen.index,
                            gen.sens_P_u_bound,
                            gen.sens_P_l_bound)))



#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

from __future__ import print_function
import time
import numpy as np
from .method_error import *
from .method import PFmethod
        
class ACOPF(PFmethod):
    """
    AC optimal power flow method.
    """

    name = 'ACOPF'

    _parameters = {'weight_cost' : 1e0,      # weight for generation cost
                   'weight_vmag' : 0.,       # weight for voltage magnitude regularization
                   'weight_vang' : 0.,       # weight for voltage angle regularization
                   'weight_pq' : 0.,         # weight for generator power regularization
                   'weight_t' : 0.,          # weight for tap ratios regularization
                   'weight_b' : 0.,          # weight for shunt susceptances regularization
                   'thermal_limits': 'none', # none, linear, nonlinear
                   'vmin_thresh': 0.1,       # threshold for vmin termination
                   'solver': 'inlp'}         # OPTALG optimization solver (augl, ipopt, inlp)

    _parameters_augl = {'feastol' : 1e-4,
                        'optol' : 1e-4,
                        'kappa' : 1e-2}

    _parameters_ipopt = {}

    _parameters_inlp = {}

    def __init__(self):

        from optalg.opt_solver import OptSolverAugL, OptSolverIpopt, OptSolverINLP
        
        # Parent init
        PFmethod.__init__(self)

        # Solver params
        augl_params = OptSolverAugL.parameters.copy()
        augl_params.update(self._parameters_augl)   # overwrite defaults

        ipopt_params = OptSolverIpopt.parameters.copy()
        ipopt_params.update(self._parameters_ipopt) # overwrite defaults

        inlp_params = OptSolverINLP.parameters.copy()
        inlp_params.update(self._parameters_inlp)   # overwrite defaults

        self._parameters = ACOPF._parameters.copy()
        self._parameters['solver_parameters'] = {'augl': augl_params,
                                                 'ipopt': ipopt_params,
                                                 'inlp': inlp_params}
                   
    def create_problem(self,net):
        
        import pfnet

        # Parameters
        params = self._parameters
        wcost  = params['weight_cost']
        wvmag  = params['weight_vmag']
        wvang = params['weight_vang']
        wpq = params['weight_pq']
        wt = params['weight_t']
        wb = params['weight_b']        
        th = params['thermal_limits']
        
        # Clear flags
        net.clear_flags()
        
        # Voltage magnitudes
        net.set_flags('bus',
                      ['variable','bounded'], 
                      'any',
                      'voltage magnitude')
        
        # Voltage angles
        net.set_flags('bus',
                      'variable',
                      'not slack',
                      'voltage angle')

        # Generator powers
        net.set_flags('generator',
                      ['variable','bounded'],
                      'any',
                      ['active power','reactive power'])

        try:
            assert(net.num_vars == (2*net.get_num_buses(True)-net.get_num_slack_buses(True) +
                                    2*net.get_num_generators(True))*net.num_periods)
            assert(net.num_bounded == (2*net.get_num_generators(True) +
                                       net.get_num_buses(True))*net.num_periods)
        except AssertionError:
            raise PFmethodError_BadProblem()
                                    
        # Problem
        problem = pfnet.Problem(net)

        # Constraints
        problem.add_constraint(pfnet.Constraint('AC power balance',net))
        problem.add_constraint(pfnet.Constraint('variable bounds',net))
        if th == 'nonlinear':
            problem.add_constraint(pfnet.Constraint("AC branch flow limits",net))
        elif th == 'linear':
            problem.add_constraint(pfnet.Constraint("linearized AC branch flow limits",net))
        elif th == 'none':
            pass
        else:
            raise PFmethodError_BadParamValue('thermal_limits')

        # Functions
        problem.add_function(pfnet.Function('generation cost',
                                            wcost/max([net.get_num_generators(True),1.]),net))
        if wvmag:
            problem.add_function(pfnet.Function('voltage magnitude regularization',
                                                wvmag/max([net.get_num_buses(True),1.]),net))
        if wvang:
            problem.add_function(pfnet.Function('voltage angle regularization',
                                                wvang/max([net.get_num_buses(True),1.]),net))
        if wpq:
            problem.add_function(pfnet.Function('generator powers regularization',
                                                wpq/max([net.get_num_generators(True),1.]),net))
        if wt:
            problem.add_function(pfnet.Function('tap ratio regularization',
                                                wt/max([net.get_num_tap_changers(True),1.]),net))
        if wb:
            problem.add_function(pfnet.Function('susceptance regularization',
                                                wb/max([net.get_num_switched_shunts(True),1.]),net))
        problem.analyze()
        
        # Return
        return problem
            
    def solve(self,net):

        from optalg.opt_solver import OptSolverError, OptTermination
        from optalg.opt_solver import OptSolverAugL, OptSolverIpopt, OptSolverINLP
        
        # Parameters
        params = self._parameters
        vmin_thresh = params['vmin_thresh']
        solver_name = params['solver']
        solver_params = params['solver_parameters']

        # Opt solver
        if solver_name == 'augl':
            solver = OptSolverAugL()
        elif solver_name == 'ipopt':
            solver = OptSolverIpopt()
        elif solver_name == 'inlp':
            solver = OptSolverINLP()
        else:
            raise PFmethodError_BadOptSolver()
        solver.set_parameters(solver_params[solver_name])

        # Copy network
        net = net.get_copy(merge_buses=True)
        self.set_network_snapshot(net)
        
        # Problem
        t0 = time.time()
        problem = self.create_problem(net)
        problem_time = time.time()-t0

        # Termination
        def t1(s):
            if np.min(s.problem.wrapped_problem.network.bus_v_min) < vmin_thresh:
                return True
            else:
                return False
        solver.add_termination(OptTermination(t1,'low voltage'))
                
        # Solve
        update = True
        t0 = time.time()
        try:
            solver.solve(problem)
        except OptSolverError as e:
            raise PFmethodError_SolverError(e)
        except Exception as e:
            update = False
            raise e
        finally:
            
            # Update network
            if update:
                net.set_var_values(solver.get_primal_variables()[:net.num_vars])
                net.update_properties()
                net.clear_sensitivities()
                problem.store_sensitivities(*solver.get_dual_variables())

            # Save results
            self.set_solver_name(solver_name)
            self.set_solver_status(solver.get_status())
            self.set_solver_message(solver.get_error_msg())
            self.set_solver_iterations(solver.get_iterations())
            self.set_solver_time(time.time()-t0)
            self.set_solver_primal_variables(solver.get_primal_variables())
            self.set_solver_dual_variables(solver.get_dual_variables())
            self.set_problem(None) # skip for now
            self.set_problem_time(problem_time)
            self.set_network_snapshot(net)




#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

from __future__ import print_function
import time
import numpy as np
from .method_error import *
from .method import PFmethod
from numpy.linalg import norm

class DCOPF(PFmethod):
    """
    DC optimal power flow method.
    """

    name = 'DCOPF'
        
    _parameters = {'thermal_limits': False,
                   'renewable_curtailment': False,
                   'solver' : 'iqp'}

    _parameters_iqp = {}

    _parameters_augl = {}

    _parameters_ipopt = {}
    
    def __init__(self):

        from optalg.opt_solver import OptSolverIQP, OptSolverAugL, OptSolverIpopt

        # Parent init
        PFmethod.__init__(self)
        
        # Solver parameters
        iqp_params = OptSolverIQP.parameters.copy()
        iqp_params.update(self._parameters_iqp)     # overwrite defaults

        augl_params = OptSolverAugL.parameters.copy()
        augl_params.update(self._parameters_augl)   # overwrite defaults

        ipopt_params = OptSolverIpopt.parameters.copy()
        ipopt_params.update(self._parameters_ipopt) # overwrite defaults

        self._parameters = DCOPF._parameters.copy()
        self._parameters['solver_parameters'] = {'iqp': iqp_params,
                                                 'augl': augl_params,
                                                 'ipopt': ipopt_params}

    def create_problem(self,net):

        import pfnet
        
        # Parameters
        params = self._parameters
        thermal_limits = params['thermal_limits']
        
        # Clear flags
        net.clear_flags()
        
        # Set flags
        net.set_flags('bus',
                      'variable',
                      'not slack',
                      'voltage angle')
        net.set_flags('generator',
                      ['variable','bounded'],
                      'any',
                      'active power')
        net.set_flags('load',
                      ['variable','bounded'],
                      'adjustable active power',
                      'active power')
        if params['renewable_curtailment']:
            net.set_flags('variable generator',
                          ['variable','bounded'],
                          'any',
                          'active power')

        try:
            num_gvar = net.get_num_generators()
            num_cur = net.num_var_generators if params['renewable_curtailment'] else 0
            assert(net.num_bounded == (num_gvar+net.get_num_P_adjust_loads()+num_cur)*net.num_periods)
            assert(net.num_vars == (net.num_buses-net.get_num_slack_buses()+
                                    num_gvar+net.get_num_P_adjust_loads()+
                                    num_cur)*net.num_periods)
        except AssertionError:
            raise PFmethodError_BadProblem()
            
        # Set up problem
        problem = pfnet.Problem(net)
        problem.add_constraint(pfnet.Constraint('variable bounds',net))
        problem.add_constraint(pfnet.Constraint('DC power balance',net))
        if thermal_limits:
            problem.add_constraint(pfnet.Constraint('DC branch flow limits',net))
        problem.add_function(pfnet.Function('generation cost',1.,net))
        problem.add_function(pfnet.Function('consumption utility',-1.,net))
        problem.analyze()
        
        # Return
        return problem
            
    def solve(self,net):

        from optalg.opt_solver import OptSolverError
        from optalg.opt_solver import OptSolverIQP, OptSolverAugL, OptSolverIpopt
        
        # Parameters
        params = self._parameters
        solver_name = params['solver']
        solver_params = params['solver_parameters']

        # Solver
        if solver_name == 'iqp':
            solver = OptSolverIQP()
        elif solver_name == 'augl':
            solver = OptSolverAugL()
        elif solver_name == 'ipopt':
            solver = OptSolverIpopt()
        else:
            raise PFmethodError_BadOptSolver()
        solver.set_parameters(solver_params[solver_name])

        # Copy network
        net = net.get_copy(merge_buses=True)

        # Problem
        t0 = time.time()
        problem = self.create_problem(net)
        problem_time = time.time()-t0
                
        # Solve
        update = True
        t0 = time.time()
        try:
            solver.solve(problem)
        except OptSolverError as e:
            raise PFmethodError_SolverError(e)
        except Exception as e:
            update = False
            raise e
        finally:

            # Update network
            if update:
                net.set_var_values(solver.get_primal_variables()[:net.num_vars])
                net.update_properties()
                net.clear_sensitivities()
                problem.store_sensitivities(*solver.get_dual_variables())

            # Save results
            self.set_solver_name(solver_name)
            self.set_solver_status(solver.get_status())
            self.set_solver_message(solver.get_error_msg())
            self.set_solver_iterations(solver.get_iterations())
            self.set_solver_time(time.time()-t0)
            self.set_solver_primal_variables(solver.get_primal_variables())
            self.set_solver_dual_variables(solver.get_dual_variables())
            self.set_problem(None) # skip for now
            self.set_problem_time(problem_time)
            self.set_network_snapshot(net)



#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

from __future__ import print_function
import os
import pfnet
import unittest
import numpy as np
import gridopt as gopt

class TestScripts(unittest.TestCase):
    
    def setUp(self):
                    
        pass

    def test_gridopt(self):

        try:
            
            case = os.path.join('.', 'tests', 'resources', 'cases', 'ieee300.mat')
            
            net = pfnet.Parser(case).parse(case)
            
            method = gopt.power_flow.ACPF()
            method.set_parameters({'quiet': True, 'solver': 'nr'})
            method.solve(net)
            method.update_network(net)
            
            gopt.scripts.gridopt.main([case, 'ACPF', '--parameters', 'solver=nr', '--write', 'foo.json', '--quiet'])

            p = pfnet.ParserJSON()
            net_json = p.parse('foo.json')
            pfnet.tests.utils.compare_networks(self, net, net_json)

            gopt.scripts.gridopt.main([case, 'ACPF', '--parameters', 'solver=nr', '--write', 'foo.m', '--quiet'])

            p = pfnet.PyParserMAT()
            net_m = p.parse('foo.m')
            pfnet.tests.utils.compare_networks(self, net, net_m)
            
        finally:
            if os.path.isfile('foo.json'):
                os.remove('foo.json')
            if os.path.isfile('foo.m'):
                os.remove('foo.m')
            
        



#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

from __future__ import print_function
import os
import copy
import optalg
import unittest
import numpy as np
import pfnet as pf
from . import utils
import gridopt as gopt
from numpy.linalg import norm

class TestPowerFlow(unittest.TestCase):
    
    def setUp(self):
                    
        pass

    def test_ACPF_opt_controls_support(self):

        case = os.path.join('tests', 'resources', 'cases', 'aesoSL2014.raw')
        if not os.path.isfile(case):
            raise unittest.SkipTest('file not available')

        raise unittest.SkipTest('test incomplete')

    def test_ACPF_nr_controls_support(self):

        case = os.path.join('tests', 'resources', 'cases', 'aesoSL2014.raw')
        if not os.path.isfile(case):
            raise unittest.SkipTest('file not available')

        net = pf.Parser(case).parse(case)

        # Default
        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr'})
        self.assertTrue(method._parameters['Q_limits'])
        self.assertTrue(method._parameters['tap_limits'])
        self.assertTrue(method._parameters['shunt_limits'])
        self.assertEqual(method._parameters['Q_mode'], 'regulating')
        self.assertEqual(method._parameters['tap_mode'], 'locked')
        self.assertEqual(method._parameters['shunt_mode'], 'locked')
        method.solve(net)
        self.assertEqual(method.get_results()['solver status'], 'solved')

        # Q
        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'Q_mode': 'regulating',
                               'Q_limits': False})
        method.solve(net)
        self.assertEqual(method.get_results()['solver status'], 'solved')

        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'Q_mode': 'free',
                               'Q_limits': False})
        self.assertRaises(ValueError, method.solve, net)

        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'Q_mode': 'free',
                               'Q_limits': True})
        self.assertRaises(ValueError, method.solve, net)

        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'Q_mode': 'locked',
                               'Q_limits': True})
        self.assertRaises(ValueError, method.solve, net)

        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'Q_mode': 'locked',
                               'Q_limits': False})
        self.assertRaises(ValueError, method.solve, net)
        
        # Shunts
        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'shunt_mode': 'locked',
                               'shunt_limits': False})
        method.solve(net)
        self.assertEqual(method.get_results()['solver status'], 'solved')

        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'shunt_mode': 'regulating',
                               'shunt_limits': False})
        self.assertRaises(ValueError, method.solve, net)

        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'shunt_mode': 'regulating',
                               'shunt_limits': True})
        method.solve(net)
        self.assertEqual(method.get_results()['solver status'], 'solved')

        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'shunt_mode': 'free',
                               'shunt_limits': False})
        self.assertRaises(ValueError, method.solve, net)

        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'shunt_mode': 'free',
                               'shunt_limits': True})
        self.assertRaises(ValueError, method.solve, net)

        # Taps
        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'tap_mode': 'locked',
                               'tap_limits': False})
        method.solve(net)
        self.assertEqual(method.get_results()['solver status'], 'solved')

        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'tap_mode': 'regulating',
                               'tap_limits': False})
        self.assertRaises(ValueError, method.solve, net)

        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'tap_mode': 'regulating',
                               'tap_limits': True})
        method.solve(net)
        self.assertEqual(method.get_results()['solver status'], 'solved')

        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'tap_mode': 'free',
                               'tap_limits': False})
        self.assertRaises(ValueError, method.solve, net)

        method = gopt.power_flow.ACPF()
        method.set_parameters({'quiet': True,
                               'solver': 'nr',
                               'tap_mode': 'free',
                               'tap_limits': True})
        self.assertRaises(ValueError, method.solve, net)

    def test_ACPF_keep_all(self):

        print('')

        for case in utils.test_cases:

            if os.path.splitext(case)[1] == '.raw':

                parser = pf.ParserRAW()
                net = parser.parse(case)
                
                parser.set('keep_all_out_of_service', True)
                net_oos = parser.parse(case)

                if (net.num_buses == net_oos.num_buses and
                    net.num_generators == net_oos.num_generators and
                    net.num_loads == net_oos.num_loads and
                    net.num_shunts == net_oos.num_shunts and
                    net.num_branches == net_oos.num_branches):
                    continue

                method = gopt.power_flow.ACPF()
                method.set_parameters({'solver': 'augl', 'quiet': True, 'v_mag_warm_ref': True})

                method.solve(net)
                r = method.get_results()

                method.solve(net_oos)
                r_oos = method.get_results()

                self.assertEqual(r['solver status'], r_oos['solver status'])
                self.assertEqual(r['solver iterations'], r_oos['solver iterations'])
                self.assertEqual(r['network snapshot'].bus_v_max, r_oos['network snapshot'].bus_v_max)
                self.assertEqual(r['network snapshot'].bus_v_min, r_oos['network snapshot'].bus_v_min)
                self.assertEqual(r['network snapshot'].bus_P_mis, r_oos['network snapshot'].bus_P_mis)
                self.assertEqual(r['network snapshot'].bus_Q_mis, r_oos['network snapshot'].bus_Q_mis)

                self.assertRaises(AssertionError, pf.tests.utils.compare_networks, self, r['network snapshot'], r_oos['network snapshot'])
                
    def test_ACOPF_parameters(self):

        acopf = gopt.power_flow.ACOPF()

        # Set parameters
        acopf.set_parameters({'weight_cost': 1.1,
                              'weight_vmag' : 1.2,
                              'weight_vang' : 1.3,
                              'weight_pq' : 1.4,
                              'weight_t' : 1.5,
                              'weight_b' : 1.6,
                              'thermal_limits' : True,
                              'vmin_thresh' : 0.123,
                              'solver' : 'inlp',
                              'beta_small' : 8.9,
                              'hessian_approximation' : 'test',
                              'maxiter': 432})

        # Check exception for invalid param
        self.assertRaises(gopt.power_flow.method_error.PFmethodError_BadParams,
                          acopf.set_parameters,{'foo' : 'bar'})

        # Get parameters
        params = acopf.get_parameters()

        # Check that set_parameters worked
        self.assertEqual(params['weight_cost'],1.1)
        self.assertEqual(params['weight_vmag'],1.2)
        self.assertEqual(params['weight_vang'],1.3)
        self.assertEqual(params['weight_pq'],1.4)
        self.assertEqual(params['weight_t'],1.5)
        self.assertEqual(params['weight_b'],1.6)
        self.assertEqual(params['thermal_limits'],True)
        self.assertEqual(params['vmin_thresh'],0.123)
        self.assertEqual(params['solver'],'inlp')
        self.assertEqual(params['solver_parameters']['augl']['beta_small'],8.9)
        self.assertEqual(params['solver_parameters']['augl']['maxiter'],432)
        self.assertEqual(params['solver_parameters']['inlp']['maxiter'],432)
        self.assertEqual(params['solver_parameters']['ipopt']['hessian_approximation'],'test')

        # Make a deep copy
        new_params = copy.deepcopy(params)

        # Set manually solver parametres
        new_params['solver_parameters']['augl']['beta_small'] = 1.333e-5
        new_params['solver_parameters']['inlp']['maxiter'] = 555
        new_params['solver_parameters']['ipopt']['hessian_approximation'] = 'new test'

        # Check that this is a separate params dictionary
        self.assertNotEqual(new_params['solver_parameters']['augl']['beta_small'],
                            params['solver_parameters']['augl']['beta_small'])
        self.assertNotEqual(new_params['solver_parameters']['inlp']['maxiter'],
                            params['solver_parameters']['inlp']['maxiter'])
        self.assertNotEqual(new_params['solver_parameters']['ipopt']['hessian_approximation'],
                            params['solver_parameters']['ipopt']['hessian_approximation'])

        # Test setting parameters that specify solver parameters
        acopf.set_parameters(new_params)

        # Check that setting solver parameters worked
        self.assertEqual(new_params['solver_parameters']['augl']['beta_small'],
                         params['solver_parameters']['augl']['beta_small'])
        self.assertEqual(new_params['solver_parameters']['inlp']['maxiter'],
                         params['solver_parameters']['inlp']['maxiter'])
        self.assertEqual(new_params['solver_parameters']['ipopt']['hessian_approximation'],
                         params['solver_parameters']['ipopt']['hessian_approximation'])

    def test_DCPF(self):

        for case in utils.test_cases:

            method = gopt.power_flow.new_method('DCPF')
            self.assertTrue(isinstance(method, gopt.power_flow.DCPF))

            method.set_parameters({'solver' : 'superlu'})
            self.assertTrue('solver_parameters' in method.get_parameters().keys())

            net = pf.Parser(case).parse(case)

            method.solve(net)

            results = method.get_results()

            self.assertEqual(results['solver status'], 'solved')
            self.assertTrue(results['solver name'] in ['mumps','superlu'])
            self.assertTrue(isinstance(results['network snapshot'], pf.Network))

    def test_ACPF_with_gen_outages(self):

        case = os.path.join('tests', 'resources', 'cases', 'ieee25.raw')
        if not os.path.isfile(case):
            raise unittest.SkipTest('file not available')
        
        net = pf.Parser(case).parse(case)

        method = gopt.power_flow.new_method('ACPF')
        
        for solver in ['nr', 'augl']:
            method.set_parameters(params={'solver': solver,
                                          'quiet': True})
            method.solve(net)

            self.assertEqual(method.get_results()['solver status'], 'solved')
            net1 = method.get_results()['network snapshot']
            
            for bus in net.buses:
                if bus.is_v_set_regulated():
                    reg_gens = [g.index for g in bus.reg_generators]
                    gen = bus.reg_generators[-1]
                    gen.in_service = False
                    method.solve(net)
                    self.assertEqual(method.get_results()['solver status'], 'solved')
                    net2 = method.get_results()['network snapshot']
                    self.assertEqual(net.get_generator(gen.index).Q,
                                     net2.get_generator(gen.index).Q) 
                    self.assertNotEqual(net.get_generator(gen.index).Q,
                                        net1.get_generator(gen.index).Q)
                    gen.in_service = True

    def test_ACPF_with_branch_outages(self):
            
        case = os.path.join('tests', 'resources', 'cases', 'ieee300.raw')
        if not os.path.isfile(case):
            raise unittest.SkipTest('file not available')

        net = pf.Parser(case).parse(case)

        method = gopt.power_flow.new_method('ACPF')
        
        for solver in ['nr','augl']:
            method.set_parameters(params={'solver': solver,
                                          'tap_mode': 'regulating',
                                          'quiet': True})
            method.solve(net)

            self.assertEqual(method.get_results()['solver status'], 'solved')
            net1 = method.get_results()['network snapshot']

            tested = False
            for bus in net.buses:
                if bus.is_regulated_by_tran() and bus.number == 1:
                    reg_trans = [br.index for br in bus.reg_trans]
                    tran = bus.reg_trans[-1]
                    tran.in_service = False
                    method.solve(net)
                    self.assertEqual(method.get_results()['solver status'], 'solved')
                    net2 = method.get_results()['network snapshot']
                    self.assertFalse(net.get_branch(tran.index).has_flags('variable', 'tap ratio'))
                    self.assertTrue(net1.get_branch(tran.index).has_flags('variable', 'tap ratio'))
                    self.assertFalse(net2.get_branch(tran.index).has_flags('variable', 'tap ratio'))
                    self.assertNotEqual(net1.num_vars, net2.num_vars)
                    self.assertGreater(net1.num_vars, net2.num_vars)
                    tran.in_service = True
                    tested = True
            self.assertTrue(tested)

    def test_ACPF_nr_heuristics(self):

        for case in utils.test_cases:

            net = pf.Parser(case).parse(case)
            
            method = gopt.power_flow.new_method('ACPF')
            method.set_parameters(params={'solver': 'nr',
                                          'quiet': True})
            method.solve(net)
            
            results = method.get_results()
            
            net_snap = results['network snapshot']
            self.assertLess(np.abs(net_snap.bus_P_mis), 1e-2) # MW
            self.assertLess(np.abs(net_snap.bus_Q_mis), 1e-2) # MVAr
            
            eps = 1e-4
            for bus in net_snap.buses:
                if bus.is_regulated_by_gen() and not bus.is_slack():
                    for gen in bus.reg_generators:
                        self.assertLessEqual(gen.Q, gen.Q_max+eps)
                        self.assertGreaterEqual(gen.Q, gen.Q_min-eps)
                    if np.abs(bus.v_mag-bus.v_set) < eps: # v at set
                        Qtotal = 0.
                        norm = 0.
                        for gen in bus.reg_generators:
                            if np.abs(gen.Q-gen.Q_max) > eps and np.abs(gen.Q-gen.Q_min) > eps:
                                Qtotal += gen.Q
                                norm += gen.Q_par
                        for gen in bus.reg_generators:
                            if np.abs(gen.Q-gen.Q_max) > eps and np.abs(gen.Q-gen.Q_min) > eps:
                                self.assertLess(np.abs(gen.Q-gen.Q_par*Qtotal/norm), eps)
                    else: # v not at set
                        num = 0
                        for gen in bus.reg_generators:
                            if np.abs(gen.Q-gen.Q_max) <= eps or np.abs(gen.Q-gen.Q_min) <= eps:
                                num += 1
                        self.assertGreaterEqual(num, 1)

            method.update_network(net)
            self.assertEqual(net.bus_P_mis, net_snap.bus_P_mis)
            self.assertEqual(net.bus_Q_mis, net_snap.bus_Q_mis)
            self.assertLess(np.abs(net.bus_P_mis), 1e-2) # MW
            self.assertLess(np.abs(net.bus_Q_mis), 1e-2) # MVAr
                                
    def test_ACPF_solutions(self):

        print('')
        
        T = 2

        sol_types = {'sol1': 'no--controls',
                     'sol2': 'gen-controls',
                     'sol3': 'all-controls'}
        
        for case in utils.test_cases:
            for sol in list(sol_types.keys()):
                for solver in ['nr','augl','ipopt', 'inlp']:
                    
                    method = gopt.power_flow.new_method('ACPF')
                    method.set_parameters(params={'solver': solver})
                    
                    net = pf.Parser(case).parse(case)
                    netMP = pf.Parser(case).parse(case,T)
                    
                    self.assertEqual(net.num_periods,1)
                    self.assertEqual(netMP.num_periods,T)

                    # Only small
                    if net.num_buses > 4000:
                        continue
                    
                    sol_file = utils.get_pf_solution_file(case, utils.DIR_PFSOL, sol)
                    sol_data = utils.read_pf_solution_file(sol_file)
                                          
                    # Set parameters
                    if sol == 'sol1':
                        method.set_parameters({'Q_limits': False})
                    elif sol == 'sol2':
                        pass # defaults
                    elif sol == 'sol3':
                        method.set_parameters({'tap_mode': 'regulating'})
                    else:
                        raise ValueError('invalid solution type')
                    method.set_parameters({'quiet': True})

                    bus_P_mis = net.bus_P_mis
                    try:
                        method.solve(net)
                    except ImportError:
                        continue # no ipopt
                    results = method.get_results()
                    self.assertEqual(results['solver name'], solver)
                    self.assertEqual(results['solver status'],'solved')
                    self.assertEqual(net.bus_P_mis,bus_P_mis)
                    self.assertLessEqual(results['network snapshot'].bus_P_mis,bus_P_mis)
                    method.update_network(net)

                    net_snap = results['network snapshot']
                    for bus in net_snap.buses:
                        obus = net.get_bus_from_number(bus.number)
                        self.assertEqual(bus.v_ang, obus.v_ang)
                        self.assertEqual(bus.v_mag, obus.v_mag)
                        self.assertAlmostEqual(bus.P_mismatch, net.get_bus_from_number(bus.number).P_mismatch, places=10)
                    for gen in net_snap.generators:
                        self.assertEqual(gen.P, net.get_generator(gen.index).P)
                    for load in net_snap.loads:
                        self.assertEqual(load.P, net.get_load(load.index).P)
                    for shunt in net_snap.shunts:
                        self.assertEqual(shunt.g, net.get_shunt(shunt.index).g)

                    self.assertLess(abs(net.bus_P_mis), 1e-2) # MW
                    self.assertLess(abs(net.bus_Q_mis), 1e-2) # MVAr
                    
                    self.assertAlmostEqual(results['network snapshot'].bus_P_mis, net.bus_P_mis, places=10)
                    self.assertAlmostEqual(results['network snapshot'].bus_Q_mis, net.bus_Q_mis, places=10)
                    
                    method.solve(netMP)
                    resultsMP = method.get_results()
                    self.assertEqual(resultsMP['solver status'],'solved')
                    method.update_network(netMP)

                    self.assertLess(np.max(np.abs(netMP.bus_P_mis)), 1e-2) # MW
                    self.assertLess(np.max(np.abs(netMP.bus_Q_mis)), 1e-2) # MVAr

                    self.assertLess(norm(resultsMP['network snapshot'].bus_P_mis-netMP.bus_P_mis,np.inf),1e-10)
                    self.assertLess(norm(resultsMP['network snapshot'].bus_Q_mis-netMP.bus_Q_mis,np.inf),1e-10)
                    self.assertLess(norm(resultsMP['network snapshot'].gen_P_cost-netMP.gen_P_cost,np.inf),1e-10)
                    
                    # Sol validation
                    validated = ''
                    if sol_data is not None:

                        v_mag_tol = sol_data['v_mag_tol']
                        v_ang_tol = sol_data['v_ang_tol']
                        bus_data = sol_data['bus_data']

                        counter = 0
                        v_mag_error = []
                        v_ang_error = []
                        for bus_num,val in list(bus_data.items()):
                        
                            v_mag = val['v_mag']
                            v_ang = val['v_ang']
                        
                            try:
                                busMP = netMP.get_bus_from_number(bus_num)
                                bus = net.get_bus_from_number(bus_num)
                            except pf.NetworkError:
                                continue
                                                       
                            for t in range(T):
                                v_mag_error.append(np.abs(busMP.v_mag[t]-v_mag))
                                v_ang_error.append(np.abs(busMP.v_ang[t]*180./np.pi-v_ang))
                            v_mag_error.append(np.abs(bus.v_mag-v_mag))
                            v_ang_error.append(np.abs(bus.v_ang*180./np.pi-v_ang))
                       
                            counter += 1

                        self.assertEqual(len(v_mag_error),len(v_ang_error))
                        if len(v_mag_error) > 0:
                            validated = 'validated'
                            self.assertLessEqual(np.max(v_mag_error),v_mag_tol)
                            self.assertLessEqual(np.max(v_ang_error),v_ang_tol)
                        self.assertEqual(len(v_mag_error),counter*(T+1))
                        self.assertEqual(len(v_ang_error),counter*(T+1))

                    # Show
                    print("%s\t%s\t%s\t%d\t%s" %(case.split(os.sep)[-1],
                                                 sol_types[sol],
                                                 solver,
                                                 results['solver iterations'],
                                                 validated))

    def test_ACOPF_solutions(self):

        print('')
        
        eps = 0.5 # %
        
        method_ipopt = gopt.power_flow.new_method('ACOPF')
        method_ipopt.set_parameters(params={'solver':'ipopt','quiet': True})
        method_augl = gopt.power_flow.new_method('ACOPF')
        method_augl.set_parameters(params={'solver':'augl','quiet': True, 'kappa': 1e-2, 'lam_reg': 1e-2})
        method_inlp = gopt.power_flow.new_method('ACOPF')
        method_inlp.set_parameters(params={'solver':'inlp','quiet': True})

        skipcases = ['aesoSL2014.raw','case2869.mat','case9241.mat','case32.art']
            
        for case in utils.test_cases:

            if case.split(os.sep)[-1] in skipcases:
                continue

            net = pf.Parser(case).parse(case)

            # Only small
            if net.num_buses > 3300:
                continue
            
            self.assertEqual(net.num_periods,1)

            # IPOPT
            try:
                net.update_properties()
                gen_P_cost = net.gen_P_cost
                method_ipopt.solve(net)
                has_ipopt = True
                self.assertEqual(method_ipopt.results['solver status'],'solved')
                self.assertEqual(net.gen_P_cost,gen_P_cost)
                self.assertNotEqual(method_ipopt.results['network snapshot'].gen_P_cost,gen_P_cost)
                self.assertEqual(method_ipopt.results['solver name'], 'ipopt')
                x1 = method_ipopt.get_results()['solver primal variables']
                i1 = method_ipopt.get_results()['solver iterations']
                p1 = method_ipopt.get_results()['network snapshot'].gen_P_cost
                print("%s\t%s\t%d" %(case.split(os.sep)[-1],'ipopt',i1))
            except ImportError:
                has_ipopt = False

            # INLP
            net.update_properties()
            gen_P_cost = net.gen_P_cost
            method_inlp.solve(net)
            self.assertEqual(method_inlp.results['solver status'],'solved')
            self.assertEqual(net.gen_P_cost,gen_P_cost)
            self.assertNotEqual(method_inlp.results['network snapshot'].gen_P_cost,gen_P_cost)
            self.assertEqual(method_inlp.results['solver name'], 'inlp')
            x2 = method_inlp.get_results()['solver primal variables']
            i2 = method_inlp.get_results()['solver iterations']
            p2 = method_inlp.get_results()['network snapshot'].gen_P_cost
            print("%s\t%s\t%d" %(case.split(os.sep)[-1],'inlp',i2))

            # AUGL
            net.update_properties()
            gen_P_cost = net.gen_P_cost
            method_augl.solve(net)
            self.assertEqual(method_augl.results['solver status'],'solved')
            self.assertEqual(net.gen_P_cost,gen_P_cost)
            self.assertNotEqual(method_augl.results['network snapshot'].gen_P_cost,gen_P_cost)
            self.assertEqual(method_augl.results['solver name'], 'augl')
            x3 = method_augl.get_results()['solver primal variables']
            i3 = method_augl.get_results()['solver iterations']
            p3 = method_augl.get_results()['network snapshot'].gen_P_cost
            print("%s\t%s\t%d" %(case.split(os.sep)[-1],'augl',i3))

            # Checks
            if has_ipopt:
                error = 100*(p1-p3)/abs(p3)
                self.assertLess(np.abs(error),eps)
                self.assertNotEqual(p1,p3)
            error = 100*(p2-p3)/abs(p3)
            self.assertLess(np.abs(error),eps)
            self.assertNotEqual(p2,p3)

            # Feasibility
            method_augl.update_network(net)
            self.assertLess(abs(net.bus_P_mis), 1e-2) # MW
            self.assertLess(abs(net.bus_Q_mis), 1e-2) # MVAR

    def test_DCOPF_solutions(self):

        T = 2

        infcases = ['ieee25.raw', 'ieee25.m']

        skipcases = ['case1354.mat','case2869.mat',
                     'case3375wp.mat','case9241.mat']
        
        method = gopt.power_flow.new_method('DCOPF')

        for case in utils.test_cases:

            if case.split(os.sep)[-1] in skipcases:
                continue
        
            net = pf.Parser(case).parse(case,T)

            for branch in net.branches:
                if branch.ratingA == 0:
                    branch.ratingA = 100
            
            self.assertEqual(net.num_periods,T)
            
            method.set_parameters({'quiet':True, 
                                   'tol': 1e-6,
                                   'thermal_limits': True})

            try:
                net.update_properties()
                gen_P_cost = net.gen_P_cost
                method.solve(net)
                self.assertEqual(method.results['solver status'],'solved')
                self.assertEqual(method.results['solver name'], 'iqp')
                self.assertTrue(np.all(net.gen_P_cost == gen_P_cost))
                self.assertTrue(np.all(method.results['network snapshot'].gen_P_cost != gen_P_cost))
            except gopt.power_flow.PFmethodError:
                self.assertTrue(case.split(os.sep)[-1] in infcases)
                self.assertEqual(method.results['solver status'],'error')
                
            results = method.get_results()
                
            net = results['network snapshot']
           
            self.assertLess(norm(results['network snapshot'].bus_P_mis-net.bus_P_mis,np.inf),1e-10)
            self.assertLess(norm(results['network snapshot'].bus_Q_mis-net.bus_Q_mis,np.inf),1e-10)
            self.assertLess(norm(results['network snapshot'].gen_P_cost-net.gen_P_cost,np.inf),1e-10)

            gen_P_cost0 = net.gen_P_cost
            load_P_util0 = net.load_P_util
            self.assertTupleEqual(gen_P_cost0.shape,(T,))
            self.assertTupleEqual(load_P_util0.shape,(T,))
            
            x = results['solver primal variables']
            lam0,nu0,mu0,pi0 = results['solver dual variables']

            self.assertTupleEqual(x.shape,((net.num_buses-
                                            net.get_num_slack_buses()+
                                            net.get_num_generators())*T,))
            self.assertTupleEqual(x.shape,(net.num_vars,))
            self.assertTupleEqual(lam0.shape,(net.num_buses*T,))
            self.assertTrue(nu0.size == 0)
            self.assertTupleEqual(mu0.shape,(net.num_vars+net.num_branches*T,))
            self.assertTupleEqual(pi0.shape,(net.num_vars+net.num_branches*T,))

            # Network update (vars and sensitivities)
            xx = x[:net.num_vars]
            row = 0
            for t in range(T):
                for bus in net.buses:
                    if not bus.is_slack():
                        self.assertEqual(bus.v_ang[t],xx[bus.index_v_ang[t]])
                        self.assertEqual(bus.sens_v_ang_u_bound[t],mu0[bus.index_v_ang[t]])
                        self.assertEqual(bus.sens_v_ang_l_bound[t],pi0[bus.index_v_ang[t]])
                    for branch in bus.branches_k:
                        self.assertEqual(branch.sens_P_u_bound[t],mu0[net.num_vars+row])
                        self.assertEqual(branch.sens_P_l_bound[t],pi0[net.num_vars+row])
                        row += 1
                for gen in net.generators:
                    self.assertEqual(gen.P[t],xx[gen.index_P[t]])
                    self.assertEqual(gen.sens_P_u_bound[t],mu0[gen.index_P[t]])
                    self.assertEqual(gen.sens_P_l_bound[t],pi0[gen.index_P[t]])

            # No thermal limits
            method.set_parameters({'thermal_limits':False})
            method.solve(net)
            self.assertEqual(method.results['solver status'],'solved')
            results = method.get_results()
            net = results['network snapshot']
            gen_P_cost1 = net.gen_P_cost
            load_P_util1 = net.load_P_util
            lam1,nu1,mu1,pi1 = results['solver dual variables']
            self.assertTupleEqual(mu1.shape,(net.num_vars,))
            self.assertTupleEqual(pi1.shape,(net.num_vars,))

            # Elastic loads
            for load in net.loads:
                load.P_max = load.P+1.
                load.P_min = load.P-1.
            self.assertEqual(net.get_num_P_adjust_loads(),net.num_loads)
            for load in net.loads:
                self.assertFalse(load.has_flags('variable','active power'))
                self.assertFalse(load.has_flags('bounded','active power'))
                self.assertTrue(np.all(load.P_min < load.P_max))
            method.solve(net)
            for load in net.loads:
                self.assertFalse(load.has_flags('variable','active power'))
                self.assertFalse(load.has_flags('bounded','active power'))
            results = method.get_results()
            net = results['network snapshot']
            for load in net.loads:
                self.assertTrue(load.has_flags('variable','active power'))
                self.assertTrue(load.has_flags('bounded','active power'))
            self.assertEqual(method.results['solver status'],'solved')
            self.assertTrue(np.all(net.gen_P_cost-net.load_P_util < gen_P_cost1-load_P_util1))

            x = results['solver primal variables']
            lam2,nu2,mu2,pi2 = results['solver dual variables']

            self.assertTupleEqual(x.shape,((net.get_num_P_adjust_loads()+
                                            net.num_buses-
                                            net.get_num_slack_buses()+
                                            net.get_num_generators())*net.num_periods,))
            self.assertTupleEqual(x.shape,(net.num_vars,))
            self.assertTupleEqual(lam2.shape,(net.num_buses*net.num_periods,))
            self.assertTrue(nu2.size == 0)
            self.assertTupleEqual(mu2.shape,x.shape)
            self.assertTupleEqual(pi2.shape,x.shape)

            xx = x[:net.num_vars]
            for t in range(T):
                for bus in net.buses:
                    if not bus.is_slack():
                        self.assertEqual(bus.v_ang[t],xx[bus.index_v_ang[t]])
                        self.assertEqual(bus.sens_v_ang_u_bound[t],mu2[bus.index_v_ang[t]])
                        self.assertEqual(bus.sens_v_ang_l_bound[t],pi2[bus.index_v_ang[t]])
                for gen in net.generators:
                    self.assertEqual(gen.P[t],xx[gen.index_P[t]])
                    self.assertEqual(gen.sens_P_u_bound[t],mu2[gen.index_P[t]])
                    self.assertEqual(gen.sens_P_l_bound[t],pi2[gen.index_P[t]])
                for load in net.loads:
                    if load.is_P_adjustable():
                        self.assertEqual(load.P[t],xx[load.index_P[t]])
                        self.assertEqual(load.sens_P_u_bound[t],mu2[load.index_P[t]])
                        self.assertEqual(load.sens_P_l_bound[t],pi2[load.index_P[t]])
                     
    def tearDown(self):
        
        pass




#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2019, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

import os
from setuptools import setup, find_packages

exec(open(os.path.join('gridopt', 'version.py')).read())

setup(name='GRIDOPT',
      zip_safe=False,
      version=__version__,
      description='Power Grid Optimization Library',
      url='https://github.com/ttinoco/GRIDOPT',
      author='Tomas Tinoco De Rubira',
      author_email='ttinoco5687@gmail.com',
      license='BSD 2-Clause License',
      packages=find_packages(),
      entry_points={'console_scripts': ['gridopt=gridopt.scripts.gridopt:main']},
      classifiers=['Development Status :: 5 - Production/Stable',
                   'License :: OSI Approved :: BSD License',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.6'],
      install_requires=['cython>=0.20.1',
                        'numpy>=1.11.2',
                        'scipy>=0.18.1',
                        'pfnet==1.3.5rc1',
                        'optalg==1.1.8rc1',
                        'nose'])



#*****************************************************#
# This file is part of GRIDOPT.                       #
#                                                     #
# Copyright (c) 2015, Tomas Tinoco De Rubira.         #
#                                                     #
# GRIDOPT is released under the BSD 2-clause license. #
#*****************************************************#

import numpy as np
from .method_error import *

class PFmethod:

    def __init__(self):
        """
        Power flow method class.
        """
        
        self._parameters = {}
        
        self.results = {'solver name': 'unknown',
                        'solver status': 'unknown',
                        'solver message': 'none',
                        'solver iterations': 0,
                        'solver time': 0.,
                        'solver primal variables': None,
                        'solver dual variables': None,
                        'problem' : None,
                        'problem time' : 0.,
                        'network snapshot' : None}

    def create_problem(self,net):
        """
        Creates optimization problem.

        Parameters
        ----------
        net : |Network|

        Returns
        -------
        prob : |Problem|
        """
        
        return None

    def get_info_printer(self):
        """
        Gets function for printing information
        about method progress.

        Returns
        -------
        printer : Function
        """

        return lambda solver,header: None

    def get_results(self):
        """
        Gets dictionary with results.

        Returns
        -------
        results : dict
        """

        return self.results

    def set_solver_name(self, name):
        """
        Sets solver name.

        Parameters
        ----------
        name : string
        """

        self.results['solver name'] = name

    def set_solver_status(self, status):
        """
        Sets solver status.

        Parameters
        ----------
        status : string
        """
        
        self.results['solver status'] = status

    def set_solver_message(self, msg):
        """
        Sets solver message.

        Parameters
        ----------
        msg : string
        """
        
        self.results['solver message'] = msg

    def set_solver_iterations(self, k):
        """
        Sets solver iterations.

        Parameters
        ----------
        k : int
        """
        
        self.results['solver iterations'] = k

    def set_solver_time(self, t):
        """
        Sets solver time in seconds.

        Parameters
        ----------
        t : float
        """
        
        self.results['solver time'] = t

    def set_solver_primal_variables(self, x):
        """
        Sets solver primal variables.

        Parameters
        ----------
        x : vector
        """
        
        self.results['solver primal variables'] = x

    def set_solver_dual_variables(self, d):
        """
        Sets solver dual variables.

        Parameters
        ----------
        d : list
        """
        
        self.results['solver dual variables'] = d

    def set_problem(self, p):
        """
        Sets problem.
         
        Parameters
        ----------
        p : |Problem|
        """

        self.results['problem'] = p

    def set_problem_time(self, t):
        """
        Sets problem construction time in seconds.

        Parameters
        ----------
        t : float
        """
        
        self.results['problem time'] = t

    def set_network_snapshot(self, net):
        """
        Sets network snapshot.

        Parameters
        ----------
        net : |Network|
        """
        
        self.results['network snapshot'] = net

    def get_parameters(self):
        """
        Gets method parameters.

        Returns
        -------
        params : dict
        """

        return self._parameters

    def set_parameters(self,params=None,strparams=None):
        """
        Sets method parameters.

        Parameters
        ----------
        params : dict
                 Name-value pairs
        strparams: dict
                   Name-value pairs where value is a string
        """

        invalid_params = []

        SOLVER_PARAMS = 'solver_parameters'

        # List of method/solver parameter dictionaries
        dict_list = [self._parameters]
        if SOLVER_PARAMS in self._parameters:
            dict_list += list(self._parameters[SOLVER_PARAMS].values())
            
        # Parameters
        if params:
            for key,value in list(params.items()):
                if key == SOLVER_PARAMS:
                    continue
                valid_key = False
                for parameter_dict in dict_list:
                    if key in parameter_dict:
                        valid_key = True
                        parameter_dict[key] = value
                if not valid_key:
                    invalid_params.append(key)
                    
            if SOLVER_PARAMS in params and SOLVER_PARAMS in self._parameters:
                solver_params = params[SOLVER_PARAMS]
                for solver_name in self._parameters[SOLVER_PARAMS].keys():
                    if solver_name in solver_params:
                        self._parameters[SOLVER_PARAMS][solver_name].update(solver_params[solver_name])
                
        # String-based parameters (from command-line utility)
        if strparams:
            for key,valuestr in list(strparams.items()):
                valid_key = False
                for parameter_dict in dict_list:
                    if key in parameter_dict:
                        valid_key = True
                        value = parameter_dict[key]
                        if type(value) is float:
                            new_value = float(valuestr)
                        elif type(value) is int:
                            new_value = int(valuestr)
                        elif type(value) is bool:
                            if valuestr == 'True':
                                new_value = True
                            elif valuestr == 'False':
                                new_value = False
                            else:
                                raise PFmethodError_ParamNotBool()
                        else:
                            new_value = valuestr
                        parameter_dict[key] = new_value
                if not valid_key:
                    invalid_params.append(key)

        # Invalid params
        if invalid_params:
            raise PFmethodError_BadParams(invalid_params)
                
    def set_results(self,results):
        """
        Sets method results.

        Parameters
        ----------
        results : dict
        """

        self.results = results
                
    def solve(self,net):
        """
        Solves power flow problem.
        
        Parameters
        ----------
        net : |Network|
        """        
        
        pass

    def update_network(self, net):
        """
        Updates network with results.

        Parameters
        ----------
        net : |Network|
        """

        if self.results['network snapshot'] is not None:
            net.copy_from_network(self.results['network snapshot'], merged=True)
            net.update_properties()

